import variationaldist as vd
import jax
import jax.numpy as np
import numpyro.distributions as npdist



def evolve(z_g, z_l_sub, indices, betas, params, rng_key_gen, params_fixed, log_prob_model, f):
	def U(z_l_sub, beta):
		return -1. * (beta * log_prob_model(z_g, z_l_sub, params["model"], indices, False) +
			(1. - beta) * vd.log_prob_local(vdmode_l, params["vd_l"], z_l_sub, indices, False))

	def evolve_bridges(aux, i):
		z_l_sub, rho_l_sub, w, rng_key_gen = aux
		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		beta = betas[i]
		# Update momentum
		rho_inc = mom_dist.sample(rng_key)
		rho_l_sub = params["eta"] * rho_l_sub + np.sqrt(1. - params["eta"] ** 2) * rho_inc
		# Simulate dynamics
		z_l_sub_new, rho_l_sub_new = leapfrog(z_l_sub, rho_l_sub, beta)
		# Update weight
		w = w + mom_dist.log_prob(rho_l_sub_new).sum() - mom_dist.log_prob(rho_l_sub).sum()
		aux = z_l_sub_new, rho_l_sub_new, w, rng_key_gen
		return aux, 0.

	def leapfrog(z_l_sub, rho_l_sub, beta):
		rho_l_sub = rho_l_sub - eps_sub * jax.grad(U, 0)(z_l_sub, beta) / 2.
		z_l_sub = z_l_sub + eps_sub * jax.grad(K, 0)(rho_l_sub)
		rho_l_sub = rho_l_sub - eps_sub * jax.grad(U, 0)(z_l_sub, beta) / 2.
		return z_l_sub, rho_l_sub

	dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, nbridges, lfsteps, mode_eps = params_fixed
	
	if mode_eps == "nodep_vec":
		eps_sub = params["eps"][indices, :]
	else:
		eps_sub = params["eps"]

	# Momentum distribution
	mom_dist = npdist.Normal(loc = np.zeros(z_l_sub.shape), scale = np.ones(z_l_sub.shape))
	K = lambda rho: -mom_dist.log_prob(rho).sum()
	
	# Sample initial momentum
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	rho_l_sub = mom_dist.sample(rng_key)
	
	# Evolve system
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	aux = (z_l_sub, rho_l_sub, 0, rng_key)
	
	aux, _ = jax.lax.scan(evolve_bridges, aux, np.arange(nbridges))
	z_l_sub, _, w, _ = aux
	return z_l_sub, w * f


