import jax
import jax.numpy as np


class IndexManager():
	def __init__(self, n, rng_key):
		self.n = n
		self.indices = np.arange(n)
		self.current = 0
		self.rng_key = rng_key

	def next(self, m):
		if m > self.n:
			return self.full()

		if self.current + m < self.n:
			low, high = self.current, self.current + m
			indices_out = self.indices[low:high]
			self.current += m
		
		else:
			indices_out = self.indices[self.current:]
			self.current = 0
			self.shuffle()

		return indices_out

	def shuffle(self):
		self.rng_key, _ = jax.random.split(self.rng_key)
		self.indices = jax.random.permutation(self.rng_key, self.indices)

	def full(self):
		return np.arange(self.n)




