import jax.numpy as np
import jax
import bm
import iwbm
import giwbm
from jax.experimental.optimizers import adam, sgd
from tqdm import tqdm
from index_manager import IndexManager
import argparse
import opt
from model_handler import load_model
import matplotlib.pyplot as plt
import pickle

savedir = '/mnt/nfs/work1/localuha/results/'

args_parser = argparse.ArgumentParser(description = 'Process arguments')
args_parser.add_argument('-bound_mode', type = str, default = 'UHA', help = 'UHA or IW.')
args_parser.add_argument('-model', type = str, default = 'gaussian', help = 'Model to use.')
args_parser.add_argument('-index', type = int, default = 0, help = 'Index of model.')
args_parser.add_argument('-N', type = int, default=1, help = 'Number of samples to estimate gradient at each step.')
args_parser.add_argument('-m', type = int, default=10, help = 'Subsampling.')
args_parser.add_argument('-nbridges', type = int, default=5, help = 'Number of bridging densities.')
args_parser.add_argument('-lfsteps', type = int, default=1, help = 'Number of leapfrog steps.')
args_parser.add_argument('-iters', type = int, default=10000, help = 'Number of iterations.')
args_parser.add_argument('-iters_base', type = int, default=50000, help = 'Number of iterations.')
args_parser.add_argument('-lr', type = float, default=0.001, help = 'Learning rate.')
args_parser.add_argument('-id', type = int, default=-1, help = 'ID.')
args_parser.add_argument('-seed', type = int, default=1, help = 'Random seed to use.')
args_parser.add_argument('-vdmode_g', type = int, default=1, help = 'Variational distribution mode (1, 2).')
args_parser.add_argument('-vdmode_l', type = int, default=1, help = 'Variational distribution mode (1, 2).')
args_parser.add_argument('-stl_g', type = int, default=0, help = 'STL for global variables.')
args_parser.add_argument('-stl_l', type = int, default=0, help = 'STL for local variables.')
args_parser.add_argument('-param', type = int, default=1, help = '(0, 1, or 2). 0 and 1 are factors, 2 is learn')
args_parser.add_argument('-run_cluster', type = int, default=0, help = '1: true, 0: false.')
args_parser.add_argument('-mode_eps', type = str, default='nodep_vec', help = 'nodep or nodep_vec.')
info = args_parser.parse_args()



stl_g = True if info.stl_g == 1 else 0
stl_l = True if info.stl_l == 1 else 0

log_prob_model, log_prob_model_global, log_prob_model_local_ind, dim_g, dim_l, num_l, model_params = load_model(info.model, info.index, info.param)
rng_key_gen = jax.random.PRNGKey(info.seed)

print(info, dim_g, dim_l, num_l)

if info.param == 2:
	trainable = ("vd_g", "vd_l", "eps", "eta", "mgridref_y", "model_params")
else:
	trainable = ("vd_g", "vd_l", "eps", "eta", "mgridref_y")

val_and_grad_fun = jax.jit(jax.value_and_grad(bm.compute_avg_ratio, 1), static_argnums = (2, 3, 4))
loss_fun = jax.jit(bm.compute_avg_ratio, static_argnums = (2, 3, 4))

tracker = {}



# Optimize initial
params_flat, unflatten, params_fixed = bm.initialize(dim_g=dim_g, dim_l=dim_l, num_l=num_l, vdmode_g=info.vdmode_g,
	vdmode_l=info.vdmode_l, stl_g=stl_g, stl_l=stl_l, model_params=model_params, nbridges=0, trainable=trainable)

losses, _, diverged, params_flat = opt.run(info.iters_base, info, params_flat, unflatten, params_fixed, log_prob_model, log_prob_model_global,
	log_prob_model_local_ind, val_and_grad_fun, loss_fun, IndexManager(num_l, rng_key_gen), trainable, rng_key_gen)

elbo_init = -np.mean(np.array(losses[-500:]))
print("Done optimization initial.", elbo_init)

tracker["diverged"] = False 
tracker['elbo_init'] = elbo_init
tracker['elbo_final'] = elbo_init

params_train, params_notrain = unflatten(params_flat)
params = {**params_train, **params_notrain}
# print(params["model"])
# print(params["vd_g"])
# print(params["vd_l"])
# print(params["vd_g"]["mean"])
# print(np.exp(params["vd_g"]["logdiag"]))
# print(params["vd_l"]["means"][:3,:3])
# print(np.exp(params["vd_l"]["logdiags"][:5,:5]))

if info.nbridges == 0 and info.run_cluster == 1:
	name = '%i_%i_%s.pkl' % (info.id, info.nbridges, info.model)
	with open(savedir + name, 'wb') as f:
		pickle.dump((info, tracker, diverged), f)
	print('Done saving results.')
	exit()


if info.run_cluster == 0:
	import matplotlib.pyplot as plt
	plt.figure()
	plt.plot(losses)
	plt.show()
	# exit()




# Optimize full
# trainable = ("vd_g", "vd_l")
# trainable = ("vd_g", "vd_l", "eps", "eta", "model_params")
# trainable = ("vd_g", "vd_l", "eps", "eta")
params_train, params_notrain = unflatten(params_flat)
params = {**params_train, **params_notrain}
# print(params["vd_g"])
# print(params["vd_l"])


if info.bound_mode == 'UHA':
	params_flat, unflatten, params_fixed = bm.initialize(dim_g=dim_g, dim_l=dim_l, num_l=num_l, vdmode_g=info.vdmode_g,
		vdmode_l=info.vdmode_l, stl_g=stl_g, stl_l=stl_l, nbridges=info.nbridges, trainable=trainable, mode_eps=info.mode_eps,
		vdparams_g=params["vd_g"], vdparams_l=params["vd_l"], model_params=params["model"])

elif info.bound_mode == 'IW':
	if info.param == 2:
		trainable = ("vd_g", "vd_l", "model_params")
	else:
		trainable = ("vd_g", "vd_l")

	val_and_grad_fun = jax.jit(jax.value_and_grad(iwbm.compute_avg_ratio, 1), static_argnums = (2, 3, 4, 5))
	loss_fun = jax.jit(iwbm.compute_avg_ratio, static_argnums = (2, 3, 4, 5))
	
	K = info.nbridges
	if K == 0:
		K = 1

	params_flat, unflatten, params_fixed = iwbm.initialize(dim_g=dim_g, dim_l=dim_l, num_l=num_l, vdmode_g=info.vdmode_g,
		vdmode_l=info.vdmode_l, stl_g=stl_g, stl_l=stl_l, K=K, trainable=trainable,
		vdparams_g=params["vd_g"], vdparams_l=params["vd_l"], model_params=params["model"])

elif info.bound_mode == 'IWG':
	if info.param == 2:
		trainable = ("vd_g", "vd_l", "model_params")
	else:
		trainable = ("vd_g", "vd_l")

	val_and_grad_fun = jax.jit(jax.value_and_grad(giwbm.compute_avg_ratio, 1), static_argnums = (2, 3, 4))
	loss_fun = jax.jit(giwbm.compute_avg_ratio, static_argnums = (2, 3, 4))
	
	K = info.nbridges
	if K == 0:
		K = 1

	params_flat, unflatten, params_fixed = giwbm.initialize(dim_g=dim_g, dim_l=dim_l, num_l=num_l, vdmode_g=info.vdmode_g,
		vdmode_l=info.vdmode_l, stl_g=stl_g, stl_l=stl_l, K=K, trainable=trainable,
		vdparams_g=params["vd_g"], vdparams_l=params["vd_l"], model_params=params["model"])

else:
	raise NotImplementedError("Bound mode not implemented.")


losses, losses_train, diverged, params_flat = opt.run(info.iters, info, params_flat, unflatten, params_fixed, log_prob_model, log_prob_model_global,
	log_prob_model_local_ind, val_and_grad_fun, loss_fun, IndexManager(num_l, rng_key_gen), trainable, rng_key_gen, mode=info.bound_mode)

print("Done optimization.", -np.array(losses[-500:]).mean())

params_train, params_notrain = unflatten(params_flat)
params = {**params_train, **params_notrain}
# print(params["vd_g"])
# print(params["vd_l"])



params_train, params_notrain = unflatten(params_flat)
params = {**params_train, **params_notrain}
# print(params["vd_g"])
# print(params["eps"])
# print(params["eta"])
# try:
# 	print(params_notrain["vd_l"])
# except:
# 	print(params["vd_l"])

final_elbo = -np.mean(np.array(losses[-500:]))
tracker['elbo_init'] = elbo_init
tracker['elbo_final'] = final_elbo
tracker["diverged"] = diverged

if info.run_cluster == 0:
	import matplotlib.pyplot as plt
	fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True)
	axs[0].plot(losses)
	axs[1].plot(losses_train)
	plt.show()

if info.run_cluster == 1:
	name = '%i_%i_%s.pkl' % (info.id, info.nbridges, info.model)
	with open(savedir + name, 'wb') as f:
		pickle.dump((info, tracker, diverged), f)
	print('Done saving results.')



# fix, axs = plt.subplots(2, sharex=True)
# axs[0].plot(losses)
# axs[1].plot(losses_train)
# plt.show()




