from jax import jit, grad, vmap
import jax.numpy as np
import jax
import numpyro
from jax.flatten_util import ravel_pytree
import numpyro.distributions as npdists
import models.gaussian as mgauss
import models.student_t as mst
import models.student_tIW as mstIW
import models.synthetic as msyn
import models.synthetic_IW as msynIW
import models.funnel as mf
import models.movie_lens_a as mla
import models.movie_lens_b as mlb
import models.movie_lens_bIW as mlbIW


def load_model(model, index, param):
	if model == 'gaussian':
		return mgauss.load_model()
	elif model == 'student_t':
		# return mst.load_model()
		return mstIW.load_model()
	elif "synthetic_g" in model:
		# return msyn.load_model(model, index, param)
		return msynIW.load_model(model, index, param)
	elif model == "funnel":
		return mf.load_model()
	elif "movie_lens_a" in model:
		# model is movie_lens_a?u?m?s
		# where u is number of users, m number of movies, s seed
		info = model.split("?")
		n_users = int(info[1])
		n_movies = int(info[2])
		seed = int(info[3])
		rng_key = jax.random.PRNGKey(seed)
		return mla.load_model(param, n_users, n_movies, rng_key)
	elif "movie_lens_b" in model:
		info = model.split("?")
		n_users = int(info[1])
		n_movies = int(info[2])
		seed = int(info[3])
		rng_key = jax.random.PRNGKey(seed)
		# return mlb.load_model(param, n_users, n_movies, rng_key)
		return mlbIW.load_model(param, n_users, n_movies, rng_key)
	else:
		raise NotImplementedError("Model not implemented.")





