import jax.numpy as np
import numpyro.distributions as npdist
import numpy as onp
import jax
import pickle
from .reparam_aux import apply_transform
from .datasets.ml100k.build_dataset_ml import load_data
# (n_users, n_movies, rng_key)


def load_model(param, n_users, n_movies, rng_key):
	def log_prob_local_ind(z_g, z_l, params, i):
		# z_l is a vector (dim_z)
		x_l = movie_features[i, :, :] # Matrix (n_movies, d_z)
		y_l = ratings[i, :] # Vector (n_movies)

		mean_z = z_g[:dim_l]
		logscale_z = z_g[dim_l:]

		logits = np.dot(x_l, z_l) # Vector (N_i)

		lp_z = npdist.Normal(loc = mean_z, scale = np.exp(logscale_z)).log_prob(z_l)
		lp_obs = npdist.BernoulliLogits(logits = logits).log_prob(y_l)
		return lp_z.sum() + lp_obs.sum()

	def log_prob_global(z_g, params):
		dist = npdist.Normal(loc = 0., scale = 1.)
		return dist.log_prob(z_g).sum()

	def log_prob(z_g, z_l_sub, params, indices, factor=True):
		lp = log_prob_global(z_g, params)

		f = jax.lax.cond(factor, lambda : num_l * 1. / indices.shape[0], lambda : 1.)

		lp_local = jax.vmap(log_prob_local_ind, in_axes = (None, 0, None, 0))(z_g, z_l_sub, params, indices)
		lp = lp + lp_local.sum() * f
		return lp

	def reparam(z_g, z_l_sub, params, indices, factor=True):
		log_det_jac = 0.
		params_g = params["z_g"]
		params_l_sub = params["z_l"][indices, :]

		z_g_out, log_det_jac_g = apply_transform(z_g, np.zeros(dim_g), np.ones(dim_g), params_g)
		log_det_jac += log_det_jac_g

		f = jax.lax.cond(factor, lambda : num_l * 1. / indices.shape[0], lambda : 1.)

		mean_z_aux = z_g_out[:dim_l] * np.ones((indices.shape[0], dim_l))
		scale_z_aux = np.exp(z_g_out[dim_l:]) * np.ones((indices.shape[0], dim_l))
		z_l_sub_out, log_det_jac_l_sub = apply_transform(z_l_sub, mean_z_aux, scale_z_aux, params_l_sub)
		log_det_jac += log_det_jac_l_sub * f

		return z_g_out, z_l_sub_out, log_det_jac

	def reparam_local(z_g, z_l, params, i):
		log_det_jac = 0.
		params_g = params["z_g"]
		params_l = params["z_l"][i, :]

		z_g_out, _ = reparam_global(z_g, params)

		mean_z_aux = z_g_out[:dim_l] * np.ones(dim_l)
		scale_z_aux = np.exp(z_g_out[dim_l:]) * np.ones(dim_l)
		z_l_out, log_det_jac_l = apply_transform(z_l, mean_z_aux, scale_z_aux, params_l)
		log_det_jac += log_det_jac_l

		return z_g_out, z_l_out, log_det_jac

	def reparam_global(z_g, params):
		log_det_jac = 0.
		params_g = params["z_g"]

		z_g_out, log_det_jac_g = apply_transform(z_g, np.zeros(dim_g), np.ones(dim_g), params_g)
		log_det_jac += log_det_jac_g

		return z_g_out, log_det_jac

	def log_prob_final(z_g, z_l_sub, params, indices, factor=True):
		z_g_aux, z_l_sub_aux, log_det_jac = reparam(z_g, z_l_sub, params, indices, factor)
		return log_prob(z_g_aux, z_l_sub_aux, params, indices, factor) + log_det_jac

	def log_prob_local_ind_final(z_g, z_l, params, i):
		z_g_aux, z_l_aux, log_det_jac = reparam_local(z_g, z_l, params, i)
		return log_prob_local_ind(z_g_aux, z_l_aux, params, i) + log_det_jac

	def log_prob_global_final(z_g, params):
		z_g_aux, log_det_jac = reparam_global(z_g, params)
		return log_prob_global(z_g_aux, params) + log_det_jac

	user_features, movie_features, ratings = load_data(n_users, n_movies, rng_key)
	# user_features (n_users, 3)
	# movie_features (n_users, n_movies, 18)
	# ratings (n_users, n_movies)

	num_l = n_users
	dim_l = movie_features.shape[2]

	dim_g = int(2 * dim_l)

	factor_init = 0. if param == 0 else 1.
	params = {}
	params["z_g"] = np.ones(dim_g) * factor_init
	params["z_l"] = np.ones((num_l, dim_l)) * factor_init

	return log_prob_final, log_prob_global_final, log_prob_local_ind_final, dim_g, dim_l, num_l, params



