import jax.numpy as np
import numpyro.distributions as npdist
import jax


def encode_params(means, logdiags):
	return {'means': means, 'logdiags': logdiags}

def decode_params(params):
	means, logdiags = params['means'], params['logdiags']
	return means, logdiags

def initialize(dim, num_local):
	means = np.zeros((num_local, dim))
	logdiags = np.zeros((num_local, dim))
	return encode_params(means, logdiags)

def subsample(params, indices):
	means, logdiags = decode_params(params)
	return means[indices, :], logdiags[indices, :]

def to_scale(logdiags):
	return np.exp(logdiags)
	# return (logdiag + np.sqrt(4. + logdiag * logdiag)) / 2.

def reparameterize(means, logdiags, eps):
	return to_scale(logdiags) * eps + means

# Otherwise just sample all epsilon all the time, and then select
def sample_eps(rng_key, num_local_subs, dim):
	return jax.random.normal(rng_key, shape = (num_local_subs, dim))

def sample_rep(rng_key, params, indices):
	means, logdiags = decode_params(params)
	dim = means.shape[1]
	n_sub = indices.shape[0]
	means_sub, logdiags_sub = subsample(params, indices)
	eps = sample_eps(rng_key, n_sub, dim)
	return reparameterize(means_sub, logdiags_sub, eps)

def log_prob(z_sub, params, indices, factor):
	means, _ = decode_params(params)
	num_l = means.shape[0] * 1.
	means_sub, logdiags_sub = subsample(params, indices)
	means, logdiags = decode_params(params)

	f = jax.lax.cond(factor, lambda : num_l * 1. / indices.shape[0], lambda : 1.)
	
	dist_sub = npdist.Normal(loc = means_sub, scale = to_scale(logdiags_sub))
	return dist_sub.log_prob(z_sub).sum() * f

def log_prob_ind(z_l, params, i):
	means, logdiags = decode_params(params)
	mean = means[i, :]
	logdiag = logdiags[i, :]
	dist = npdist.Normal(loc = mean, scale = to_scale(logdiag))
	return dist.log_prob(z_l).sum()









